import json
import ssl
from hashlib import md5
from io import BytesIO
from typing import Callable, get_args
from urllib.parse import urljoin

import certifi
import httpx
import pandas as pd
import zarr
from authlib.integrations.base_client.errors import InvalidTokenError, MissingTokenError
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from authlib.oauth2 import TokenAuth
from authlib.oauth2.rfc6749 import OAuth2Token
from httpx import HTTPStatusError, Response
from httpx._types import HeaderTypes, URLTypes
from loguru import logger

from polaris.benchmark import (
    BenchmarkSpecification,
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.dataset import Dataset
from polaris.evaluate import BenchmarkResults
from polaris.hub.external_auth_client import ExternalAuthClient
from polaris.hub.oauth import CachedTokenAuth
from polaris.hub.polarisfs import PolarisFileSystem
from polaris.hub.settings import PolarisHubSettings
from polaris.utils.context import ProgressIndicator, tmp_attribute_change
from polaris.utils.errors import (
    InvalidDatasetError,
    PolarisCreateArtifactError,
    PolarisHubError,
    PolarisRetrieveArtifactError,
    PolarisUnauthorizedError,
)
from polaris.utils.misc import should_verify_checksum
from polaris.utils.types import (
    AccessType,
    ChecksumStrategy,
    HubOwner,
    IOMode,
    SupportedLicenseType,
    TimeoutTypes,
    ZarrConflictResolution,
)

_HTTPX_SSL_ERROR_CODE = "[SSL: CERTIFICATE_VERIFY_FAILED]"


class PolarisHubClient(OAuth2Client):
    """
    A client for the Polaris Hub API. The Polaris Hub is a central repository of datasets, benchmarks and results.
    Visit it here: [https://polarishub.io/](https://polarishub.io/).

    Bases the [`authlib` client](https://docs.authlib.org/en/latest/client/api.html#authlib.integrations.httpx_client.OAuth2Client),
    which in turns bases the [`httpx` client](https://www.python-httpx.org/advanced/#client-instances).
    See the relevant docs to learn more about how to use these clients outside of the integration with the Polaris Hub.

    Note: Closing the client
        The client should be closed after all requests have been made. For convenience, you can also use the
        client as a context manager to automatically close the client when the context is exited. Note that once
        the client has been closed, it cannot be used anymore.

        ```python
        # Make sure to close the client once finished
        client = PolarisHubClient()
        client.get(...)
        client.close()

        # Or use the client as a context manager
        with PolarisHubClient() as client:
            client.get(...)
        ```

    Note: Interacting with artifacts owned by an organization
        Soon after being added to a new organization on Polaris, there may be a delay spanning some
        minutes where you cannot upload/download artifacts where the aforementioned organization is the owner.
        If this occurs, please re-login via `polaris login --overwrite` and try again.

    Info: Async Client
        `authlib` also supports an [async client](https://docs.authlib.org/en/latest/client/httpx.html#async-oauth-2-0).
        Since we don't expect to make multiple requests to the Hub in parallel
        and due to the added complexity stemming from using the Python asyncio API,
        we are sticking to the sync client - at least for now.
    """

    def __init__(
        self,
        settings: PolarisHubSettings | None = None,
        cache_auth_token: bool = True,
        **kwargs: dict,
    ):
        """
        Args:
            settings: A `PolarisHubSettings` instance.
            cache_auth_token: Whether to cache the auth token to a file.
            **kwargs: Additional keyword arguments passed to the authlib `OAuth2Client` constructor.
        """
        self.settings = PolarisHubSettings() if settings is None else settings

        # We cache the auth token by default, but allow the user to disable this.
        self.token_auth_class = CachedTokenAuth if cache_auth_token else TokenAuth

        super().__init__(
            # OAuth2Client
            token_endpoint=self.settings.hub_token_url,
            token_endpoint_auth_method="none",
            grant_type="urn:ietf:params:oauth:grant-type:token-exchange",
            # httpx.Client
            base_url=self.settings.api_url,
            cert=self.settings.ca_bundle,
            timeout=self.settings.default_timeout,
            # Extra
            **kwargs,
        )

        # We use an external client to get an auth token that can be exchanged for a Polaris Hub token
        self.external_client = ExternalAuthClient(
            settings=self.settings, cache_auth_token=cache_auth_token, **kwargs
        )

    def _prepare_token_endpoint_body(self, body, grant_type, **kwargs):
        """
        Override to support required fields for the token exchange grant type.
        See https://datatracker.ietf.org/doc/html/rfc8693#name-request
        """
        if grant_type == "urn:ietf:params:oauth:grant-type:token-exchange":
            kwargs.update(
                {
                    "subject_token": self.external_client.token["access_token"],
                    "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "requested_token_type": "urn:ietf:params:oauth:token-type:jwt",
                }
            )
        return super()._prepare_token_endpoint_body(body, grant_type, **kwargs)

    def ensure_active_token(self, token: OAuth2Token) -> bool:
        """
        Override the active check to trigger a refetch of the token if it is not active.
        """
        is_active = super().ensure_active_token(token)
        if is_active:
            return True

        # Check if external token is still valid
        if not self.external_client.ensure_active_token(self.external_client.token):
            return False

        # If so, use it to get a new Hub token
        self.token = self.fetch_token()
        return True

    def _load_from_signed_url(self, url: URLTypes, load_fn: Callable, headers: HeaderTypes | None = None):
        """Utility function to load a file from a signed URL"""
        response = self.get(url, auth=None, headers=headers)  # type: ignore
        response.raise_for_status()
        content = BytesIO(response.content)
        return load_fn(content)

    def _base_request_to_hub(self, url: str, method: str, **kwargs):
        """Utility function since most API methods follow the same pattern"""
        response = self.request(url=url, method=method, **kwargs)

        try:
            response.raise_for_status()

        except HTTPStatusError as error:
            response_status_code = response.status_code

            # With an internal server error, we are not sure the custom error-handling code on the hub is reached.
            if response_status_code == 500:
                raise

            # If JSON is included in the response body, we retrieve it and format it for output. If not, we fallback to
            # retrieving plain text from the body. This is important for handling certain errors thrown from the backend
            # which do not contain JSON in the response.
            try:
                response = response.json()
                response = json.dumps(response, indent=2, sort_keys=True)
            except (json.JSONDecodeError, TypeError):
                response = response.text

            # The below two error cases can happen due to the JWT token containing outdated information.
            # We therefore throw a custom error with a recommended next step.
            if response_status_code == 403:
                # This happens when trying to create an artifact for an owner the user has no access to.
                raise PolarisCreateArtifactError(response=response) from error

            if response_status_code == 404:
                # This happens when an artifact doesn't exist _or_ when the user has no access to that artifact.
                raise PolarisRetrieveArtifactError(response=response) from error

            raise PolarisHubError(response=response) from error
        # Convert the response to json format if the response contains a 'text' body
        try:
            response = response.json()
        except json.JSONDecodeError:
            pass

        return response

    def get_metadata_from_response(self, response: Response, key: str) -> str | None:
        """Get custom metadata saved to the R2 object from the headers."""
        key = f"{self.settings.custom_metadata_prefix}{key}"
        return response.headers.get(key)

    def request(self, method, url, withhold_token=False, auth=httpx.USE_CLIENT_DEFAULT, **kwargs):
        """Wraps the base request method to handle errors"""
        try:
            return super().request(method, url, withhold_token, auth, **kwargs)
        except httpx.ConnectError as error:
            # NOTE (cwognum): In the stack-trace, the more specific SSLCertVerificationError is raised.
            #   We could use the traceback module to cactch this specific error, but that would be overkill here.
            if _HTTPX_SSL_ERROR_CODE in str(error):
                raise ssl.SSLCertVerificationError(
                    "We could not verify the SSL certificate. "
                    f"Please ensure the installed version ({certifi.__version__}) of the `certifi` package is the latest. "
                    "If you require the usage of a custom CA bundle, you can set the POLARIS_CA_BUNDLE "
                    "environment variable to the path of your CA bundle. For debugging, you can temporarily disable "
                    "SSL verification by setting the POLARIS_CA_BUNDLE environment variable to `false`."
                ) from error
            raise error
        except (MissingTokenError, InvalidTokenError, httpx.HTTPStatusError, OAuthError) as error:
            if isinstance(error, httpx.HTTPStatusError) and error.response.status_code != 401:
                raise
            raise PolarisUnauthorizedError(response=error.response) from error

    def login(self, overwrite: bool = False, auto_open_browser: bool = True):
        """Login to the Polaris Hub using the OAuth2 protocol.

        Warning: Headless authentication
            It is currently not possible to login to the Polaris Hub without a browser.
            See [this Github issue](https://github.com/polaris-hub/polaris/issues/30) for more info.

        Args:
            overwrite: Whether to overwrite the current token if the user is already logged in.
            auto_open_browser: Whether to automatically open the browser to visit the authorization URL.
        """
        if overwrite or self.token is None or not self.ensure_active_token(self.token):
            self.external_client.interactive_login(overwrite=overwrite, auto_open_browser=auto_open_browser)
            self.token = self.fetch_token()

        logger.success("You are successfully logged in to the Polaris Hub.")

    # =========================
    #     API Endpoints
    # =========================

    def list_datasets(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available datasets on the Polaris Hub.

        Args:
            limit: The maximum number of datasets to return.
            offset: The offset from which to start returning datasets.

        Returns:
            A list of dataset names in the format `owner/dataset_name`.
        """
        with ProgressIndicator(
            start_msg="Fetching datasets...",
            success_msg="Fetched datasets.",
            error_msg="Failed to fetch datasets.",
        ):
            response = self._base_request_to_hub(
                url="/dataset", method="GET", params={"limit": limit, "offset": offset}
            )
            dataset_list = [bm["artifactId"] for bm in response["data"]]

            return dataset_list

    def get_dataset(
        self,
        owner: str | HubOwner,
        name: str,
        verify_checksum: ChecksumStrategy = "verify_unless_zarr",
    ) -> Dataset:
        """Load a dataset from the Polaris Hub.

        Args:
            owner: The owner of the dataset. Can be either a user or organization from the Polaris Hub.
            name: The name of the dataset.
            verify_checksum: Whether to use the checksum to verify the integrity of the dataset. If None,
                will infer a practical default based on the dataset's storage location.

        Returns:
            A `Dataset` instance, if it exists.
        """

        with ProgressIndicator(
            start_msg="Fetching dataset...",
            success_msg="Fetched dataset.",
            error_msg="Failed to fetch dataset.",
        ):
            response = self._base_request_to_hub(url=f"/dataset/{owner}/{name}", method="GET")
            storage_response = self.get(response["tableContent"]["url"])

            # This should be a 307 redirect with the signed URL
            if storage_response.status_code != 307:
                try:
                    storage_response.raise_for_status()
                except HTTPStatusError as error:
                    raise PolarisHubError(
                        message="Could not get signed URL from Polaris Hub.", response=storage_response
                    ) from error

            storage_response = storage_response.json()
            url = storage_response["url"]
            headers = storage_response["headers"]

            response["table"] = self._load_from_signed_url(url=url, headers=headers, load_fn=pd.read_parquet)

            dataset = Dataset(**response)

            if should_verify_checksum(verify_checksum, dataset):
                dataset.verify_checksum()
            else:
                dataset.md5sum = response["md5Sum"]

            return dataset

    def open_zarr_file(
        self, owner: str | HubOwner, name: str, path: str, mode: IOMode, as_consolidated: bool = True
    ) -> zarr.hierarchy.Group:
        """Open a Zarr file from a Polaris dataset

        Args:
            owner: Which Hub user or organization owns the artifact.
            name: Name of the dataset.
            path: Path to the Zarr file within the dataset.
            mode: The mode in which the file is opened.
            as_consolidated: Whether to open the store with consolidated metadata for optimized reading.
                This is only applicable in 'r' and 'r+' modes.

        Returns:
            The Zarr object representing the dataset.
        """
        if as_consolidated and mode not in ["r", "r+"]:
            raise ValueError("Consolidated archives can only be used with 'r' or 'r+' mode.")

        polaris_fs = PolarisFileSystem(
            polaris_client=self,
            dataset_owner=owner,
            dataset_name=name,
        )

        try:
            store = zarr.storage.FSStore(path, fs=polaris_fs)
            if mode in ["r", "r+"] and as_consolidated:
                return zarr.open_consolidated(store, mode=mode)
            return zarr.open(store, mode=mode)

        except HTTPStatusError as error:
            # In this case, we can pass the response to provide more information
            raise PolarisHubError(message="Error opening Zarr store", response=error.response) from error
        except Exception as error:
            raise PolarisHubError(message="Error opening Zarr store") from error

    def list_benchmarks(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available benchmarks on the Polaris Hub.

        Args:
            limit: The maximum number of benchmarks to return.
            offset: The offset from which to start returning benchmarks.

        Returns:
            A list of benchmark names in the format `owner/benchmark_name`.
        """
        with ProgressIndicator(
            start_msg="Fetching benchmarks...",
            success_msg="Fetched benchmarks.",
            error_msg="Failed to fetch benchmarks.",
        ):
            # TODO (cwognum): What to do with pagination, i.e. limit and offset?
            response = self._base_request_to_hub(
                url="/benchmark", method="GET", params={"limit": limit, "offset": offset}
            )
            benchmarks_list = [f"{HubOwner(**bm['owner'])}/{bm['name']}" for bm in response["data"]]

            return benchmarks_list

    def get_benchmark(
        self,
        owner: str | HubOwner,
        name: str,
        verify_checksum: ChecksumStrategy = "verify_unless_zarr",
    ) -> BenchmarkSpecification:
        """Load a benchmark from the Polaris Hub.

        Args:
            owner: The owner of the benchmark. Can be either a user or organization from the Polaris Hub.
            name: The name of the benchmark.
            verify_checksum: Whether to use the checksum to verify the integrity of the benchmark.

        Returns:
            A `BenchmarkSpecification` instance, if it exists.
        """
        with ProgressIndicator(
            start_msg="Fetching benchmark...",
            success_msg="Fetched benchmark.",
            error_msg="Failed to fetch benchmark.",
        ):
            response = self._base_request_to_hub(url=f"/benchmark/{owner}/{name}", method="GET")

            # TODO (jstlaurent): response["dataset"]["artifactId"] is the owner/name unique identifier,
            #  but we'd need to change the signature of get_dataset to use it
            response["dataset"] = self.get_dataset(
                response["dataset"]["owner"]["slug"],
                response["dataset"]["name"],
                verify_checksum=verify_checksum,
            )

            # TODO (cwognum): As we get more complicated benchmarks, how do we still find the right subclass?
            #  Maybe through structural pattern matching, introduced in Py3.10, or Pydantic's discriminated unions?
            benchmark_cls = (
                SingleTaskBenchmarkSpecification
                if len(response["targetCols"]) == 1
                else MultiTaskBenchmarkSpecification
            )

            benchmark = benchmark_cls(**response)

            if should_verify_checksum(verify_checksum, benchmark.dataset):
                benchmark.verify_checksum()
            else:
                benchmark.md5sum = response["md5Sum"]

            return benchmark

    def upload_results(
        self,
        results: BenchmarkResults,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
    ):
        """Upload the results to the Polaris Hub.

        Info: Owner
            The owner of the results will automatically be inferred by the hub from the user making the request.
            Contrary to other artifact types, an organization cannot own a set of results.
            However, you can specify the `BenchmarkResults.contributors` field to share credit with other hub users.

        Note: Required meta-data
            The Polaris client and hub maintain different requirements as to which meta-data is required.
            The requirements by the hub are stricter, so when uploading to the hub you might
            get some errors on missing meta-data. Make sure to fill-in as much of the meta-data as possible
            before uploading.

        Note: Benchmark name and owner
            Importantly, `results.benchmark_name` and `results.benchmark_owner` must be specified
            and match an existing benchmark on the Polaris Hub. If these results were generated by
            `benchmark.evaluate(...)`, this is done automatically.

        Args:
            results: The results to upload.
            access: Grant public or private access to result
            owner: Which Hub user or organization owns the artifact. Takes precedence over `results.owner`.
        """
        with ProgressIndicator(
            start_msg="Uploading result...",
            success_msg="Uploaded result.",
            error_msg="Failed to upload result.",
        ) as progress_indicator:
            # Get the serialized model data-structure
            results.owner = HubOwner.normalize(owner or results.owner)
            result_json = results.model_dump(by_alias=True, exclude_none=True)

            # Make a request to the hub
            response = self._base_request_to_hub(
                url="/result", method="POST", json={"access": access, **result_json}
            )

            # Inform the user about where to find their newly created artifact.
            result_url = urljoin(
                self.settings.hub_url,
                f"benchmarks/{results.benchmark_owner}/{results.benchmark_name}/{response['id']}",
            )

            progress_indicator.update_success_msg(
                f"Your result has been successfully uploaded to the Hub. View it here: {result_url}"
            )

            return response

    def upload_dataset(
        self,
        dataset: Dataset,
        access: AccessType = "private",
        timeout: TimeoutTypes = (10, 200),
        owner: HubOwner | str | None = None,
        if_exists: ZarrConflictResolution = "replace",
    ):
        """Upload the dataset to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the dataset data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from just the logged-in user.

        Note: Required meta-data
            The Polaris client and hub maintain different requirements as to which meta-data is required.
            The requirements by the hub are stricter, so when uploading to the hub you might
            get some errors on missing meta-data. Make sure to fill-in as much of the meta-data as possible
            before uploading.

        Args:
            dataset: The dataset to upload.
            access: Grant public or private access to result
            timeout: Request timeout values. User can modify the value when uploading large dataset as needed.
                This can be a single value with the timeout in seconds for all IO operations, or a more granular
                tuple with (connect_timeout, write_timeout). The type of the the timout parameter comes from `httpx`.
                Since datasets can get large, it might be needed to increase the write timeout for larger datasets.
                See also: https://www.python-httpx.org/advanced/#timeout-configuration
            owner: Which Hub user or organization owns the artifact. Takes precedence over `dataset.owner`.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.
        """
        with ProgressIndicator(
            start_msg="Uploading dataset...",
            success_msg="Uploaded dataset.",
            error_msg="Failed to upload dataset.",
        ) as progress_indicator:
            # Check if a dataset license was specified prior to upload
            if not dataset.license:
                raise InvalidDatasetError(
                    f"\nPlease specify a supported license for this dataset prior to uploading to the Polaris Hub.\nOnly some licenses are supported - {get_args(SupportedLicenseType)}."
                )

            # Normalize timeout
            if timeout is None:
                timeout = self.settings.default_timeout

            # Get the serialized data-model
            # We exclude the table as it handled separately and we exclude the cache_dir as it is user-specific
            dataset.owner = HubOwner.normalize(owner or dataset.owner)
            dataset_json = dataset.model_dump(
                exclude={"cache_dir", "table"}, exclude_none=True, by_alias=True
            )

            # If the dataset uses Zarr, we will save the Zarr archive to the Hub as well
            if dataset.uses_zarr:
                dataset_json["zarrRootPath"] = f"{PolarisFileSystem.protocol}://data.zarr"

            # Uploading a dataset is a three-step process.
            # 1. Upload the dataset meta data to the hub and prepare the hub to receive the data
            # 2. Upload the parquet file to the hub
            # 3. Upload the associated Zarr archive
            # TODO: Revert step 1 in case step 2 fails - Is this needed? Or should this be taken care of by the hub?

            # Prepare the parquet file
            buffer = BytesIO()
            dataset.table.to_parquet(buffer, engine="auto")
            parquet_size = len(buffer.getbuffer())
            parquet_md5 = md5(buffer.getbuffer()).hexdigest()

            # Step 1: Upload meta-data
            # Instead of directly uploading the data, we announce to the hub that we intend to upload it.
            # We do so separately for the Zarr archive and Parquet file.
            url = f"/dataset/{dataset.owner}/{dataset.name}"
            response = self._base_request_to_hub(
                url=url,
                method="PUT",
                json={
                    "tableContent": {
                        "size": parquet_size,
                        "fileType": "parquet",
                        "md5Sum": parquet_md5,
                    },
                    "zarrContent": [md5sum.model_dump() for md5sum in dataset._zarr_md5sum_manifest],
                    "access": access,
                    **dataset_json,
                },
                timeout=timeout,
            )

            # Step 2: Upload the parquet file
            # create an empty PUT request to get the table content URL from cloudflare
            hub_response = self.request(
                url=response["tableContent"]["url"],
                method="PUT",
                headers={
                    "Content-type": "application/vnd.apache.parquet",
                },
                timeout=timeout,
            )

            if hub_response.status_code == 307:
                # If the hub returns a 307 redirect, we need to follow it to get the signed URL
                hub_response_body = hub_response.json()

                # Upload the data to the cloudflare url
                bucket_response = self.request(
                    url=hub_response_body["url"],
                    method=hub_response_body["method"],
                    headers={
                        "Content-type": "application/vnd.apache.parquet",
                        **hub_response_body["headers"],
                    },
                    content=buffer.getvalue(),
                    auth=None,
                    timeout=timeout,  # required for large size dataset
                )
                bucket_response.raise_for_status()

            else:
                hub_response.raise_for_status()

            # Step 3: Upload any associated Zarr archive
            if dataset.uses_zarr:
                with tmp_attribute_change(self.settings, "default_timeout", timeout):
                    # Copy the Zarr archive to the hub
                    dest = self.open_zarr_file(
                        owner=dataset.owner,
                        name=dataset.name,
                        path=dataset_json["zarrRootPath"],
                        mode="w",
                        as_consolidated=False,
                    )

                    # Locally consolidate Zarr archive metadata. Future updates on handling consolidated
                    # metadata based on Zarr developers' recommendations can be tracked at:
                    # https://github.com/zarr-developers/zarr-python/issues/1731
                    zarr.consolidate_metadata(dataset.zarr_root.store.store)
                    zmetadata_content = dataset.zarr_root.store.store[".zmetadata"]
                    dest.store[".zmetadata"] = zmetadata_content

                    logger.info("Copying Zarr archive to the Hub. This may take a while.")
                    zarr.copy_store(
                        source=dataset.zarr_root.store.store,
                        dest=dest.store,
                        log=logger.debug,
                        if_exists=if_exists,
                    )

            progress_indicator.update_success_msg(
                "Your dataset has been successfully uploaded to the Hub. "
                f"View it here: {urljoin(self.settings.hub_url, f'datasets/{dataset.owner}/{dataset.name}')}"
            )

            return response

    def upload_benchmark(
        self,
        benchmark: BenchmarkSpecification,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
    ):
        """Upload the benchmark to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the benchmark data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from the logged-in user.

        Note: Required meta-data
            The Polaris client and hub maintain different requirements as to which meta-data is required.
            The requirements by the hub are stricter, so when uploading to the hub you might
            get some errors on missing meta-data. Make sure to fill-in as much of the meta-data as possible
            before uploading.

        Note: Non-existent datasets
            The client will _not_ upload the associated dataset to the hub if it does not yet exist.
            Make sure to specify an existing dataset or upload the dataset first.

        Args:
            benchmark: The benchmark to upload.
            access: Grant public or private access to result
            owner: Which Hub user or organization owns the artifact. Takes precedence over `benchmark.owner`.
        """
        with ProgressIndicator(
            start_msg="Uploading benchmark...",
            success_msg="Uploaded benchmark.",
            error_msg="Failed to upload benchmark.",
        ) as progress_indicator:
            # Get the serialized data-model
            # We exclude the dataset as we expect it to exist on the hub already.
            benchmark.owner = HubOwner.normalize(owner or benchmark.owner)
            benchmark_json = benchmark.model_dump(exclude={"dataset"}, exclude_none=True, by_alias=True)
            benchmark_json["datasetArtifactId"] = benchmark.dataset.artifact_id
            benchmark_json["access"] = access

            url = f"/benchmark/{benchmark.owner}/{benchmark.name}"
            response = self._base_request_to_hub(url=url, method="PUT", json=benchmark_json)

            progress_indicator.update_success_msg(
                "Your benchmark has been successfully uploaded to the Hub. "
                f"View it here: {urljoin(self.settings.hub_url, f'benchmarks/{benchmark.owner}/{benchmark.name}')}"
            )

            return response



import abc
import re

from loguru import logger
from pydantic import BaseModel, PrivateAttr, computed_field

from polaris.utils.errors import PolarisChecksumError


class ChecksumMixin(BaseModel, abc.ABC):
    """
    Mixin class to add checksum functionality to a class.
    """

    _md5sum: str | None = PrivateAttr(None)

    @abc.abstractmethod
    def _compute_checksum(self) -> str:
        """Compute the checksum of the dataset."""
        raise NotImplementedError

    @computed_field
    @property
    def md5sum(self) -> str:
        """Lazily compute the checksum once needed."""
        if not self.has_md5sum:
            logger.info("Computing the checksum. This can be slow for large datasets.")
            self.md5sum = self._compute_checksum()
        return self._md5sum

    @md5sum.setter
    def md5sum(self, value: str):
        """Set the checksum."""
        if not re.fullmatch(r"^[a-f0-9]{32}$", value):
            raise ValueError("The checksum should be the 32-character hexdigest of a 128 bit MD5 hash.")
        self._md5sum = value

    @property
    def has_md5sum(self) -> bool:
        """Whether the md5sum for this class has been computed and stored."""
        return self._md5sum is not None

    def verify_checksum(self, md5sum: str | None = None):
        """
        Recomputes the checksum and verifies whether it matches the stored checksum.

        Warning: Slow operation
            This operation can be slow for large datasets.

        Info: Only works for locally stored datasets
            The checksum verification only works for datasets that are stored locally in its entirety.
            We don't have to verify the checksum for datasets stored on the Hub, as the Hub will do this on upload.
            And if you're streaming the data from the Hub, we will check the checksum of each chunk on download.
        """
        if md5sum is None:
            md5sum = self._md5sum
        if md5sum is None:
            logger.warning(
                "No checksum to verify against. Specify either the md5sum parameter or "
                "store the checksum in the dataset.md5sum attribute."
            )
            return

        # Recompute the checksum
        logger.info("To verify the checksum, we need to recompute it. This can be slow for large datasets.")
        self.md5sum = self._compute_checksum()

        if self.md5sum != md5sum:
            raise PolarisChecksumError(
                f"The specified checksum {md5sum} does not match the computed checksum {self.md5sum}"
            )



from httpx import Response

from polaris.mixins._format_text import FormattingMixin  # Imported with full path to avoid circular import


class InvalidDatasetError(ValueError):
    pass


class InvalidBenchmarkError(ValueError):
    pass


class InvalidResultError(ValueError):
    pass


class TestAccessError(Exception):
    # Prevent pytest to collect this as a test
    __test__ = False

    pass


class PolarisChecksumError(ValueError):
    pass


class InvalidZarrChecksum(Exception):
    pass


class PolarisHubError(Exception, FormattingMixin):
    def __init__(self, message: str = "", response: Response | None = None):
        prefix = "The request to the Polaris Hub failed."

        if response is not None:
            prefix += f" The Hub responded with:\n{response}"

        super().__init__("\n".join([prefix, message]))


class PolarisUnauthorizedError(PolarisHubError):
    def __init__(self, response: Response | None = None):
        message = (
            "You are not logged in to Polaris or your login has expired. "
            "You can use the Polaris CLI to easily authenticate yourself again with `polaris login --overwrite`."
        )
        message = self.format(message, [self.BOLD, self.YELLOW])
        super().__init__(message, response)


class PolarisCreateArtifactError(PolarisHubError):
    def __init__(self, response: Response | None = None):
        message = (
            "Note: If you can confirm that you are authorized to perform this action, "
            "please call 'polaris login --overwrite' and try again. If the issue persists, please reach out to the Polaris team for support."
        )
        message = self.format(message, [self.BOLD, self.YELLOW])
        super().__init__(message, response)


class PolarisRetrieveArtifactError(PolarisHubError):
    def __init__(self, response: Response | None = None):
        message = (
            "Note: If this artifact exists and you can confirm that you are authorized to retrieve it, "
            "please call 'polaris login --overwrite' and try again. If the issue persists, please reach out to the Polaris team for support."
        )
        message = self.format(message, [self.BOLD, self.YELLOW])
        super().__init__(message, response)



import os
import pytest
import numpy as np
import pandas as pd

import polaris as po
from polaris.benchmark import (
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.evaluate._metric import Metric
from polaris.evaluate._results import BenchmarkResults
from polaris.utils.types import HubOwner
from polaris.dataset import Dataset


def test_result_to_json(tmpdir: str, test_user_owner: HubOwner):
    scores = pd.DataFrame(
        {
            "Test set": ["A", "A", "A", "A", "B", "B", "B", "B"],
            "Target label": ["C", "C", "D", "D", "C", "C", "D", "D"],
            "Metric": ["mean_absolute_error", "accuracy"] * 4,
            "Score": [0.1] * 8,
        }
    )

    result = BenchmarkResults(
        name="test",
        description="Lorem ipsum!",
        tags=["test"],
        user_attributes={"key": "value"},
        owner=test_user_owner,
        results=scores,
        benchmark_name="my-benchmark",
        benchmark_owner=test_user_owner,
        github_url="https://github.com/",
        paper_url="https://chemrxiv.org/",
        contributors=["my-user", "other-user"],
    )

    path = os.path.join(tmpdir, "result.json")
    result.to_json(path)
    BenchmarkResults.from_json(path)
    assert po.__version__ == result.polaris_version


def test_metrics_singletask_reg(tmpdir: str, test_single_task_benchmark: SingleTaskBenchmarkSpecification):
    _, test = test_single_task_benchmark.get_train_test_split()
    predictions = np.random.random(size=test.inputs.shape[0])
    result = test_single_task_benchmark.evaluate(predictions)
    assert isinstance(result.results, pd.DataFrame)
    assert set(result.results.columns) == {
        "Test set",
        "Target label",
        "Metric",
        "Score",
    }
    for metric in test_single_task_benchmark.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_multitask_reg(tmpdir: str, test_multi_task_benchmark: MultiTaskBenchmarkSpecification):
    train, test = test_multi_task_benchmark.get_train_test_split()
    predictions = {
        target_col: np.random.random(size=test.inputs.shape[0]) for target_col in train.target_cols
    }
    result = test_multi_task_benchmark.evaluate(predictions)
    for metric in test_multi_task_benchmark.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_singletask_clf(
    tmpdir: str, test_single_task_benchmark_clf: SingleTaskBenchmarkSpecification
):
    _, test = test_single_task_benchmark_clf.get_train_test_split()
    predictions = np.random.randint(2, size=test.inputs.shape[0])
    probabilities = np.random.uniform(size=test.inputs.shape[0])
    result = test_single_task_benchmark_clf.evaluate(y_pred=predictions, y_prob=probabilities)
    for metric in test_single_task_benchmark_clf.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_singletask_multicls_clf(
    tmpdir: str, test_single_task_benchmark_multi_clf: SingleTaskBenchmarkSpecification
):
    _, test = test_single_task_benchmark_multi_clf.get_train_test_split()
    predictions = np.random.randint(3, size=test.inputs.shape[0])
    probablities = np.random.random(size=(test.inputs.shape[0], 3))
    probablities = probablities / probablities.sum(axis=1, keepdims=True)
    result = test_single_task_benchmark_multi_clf.evaluate(y_pred=predictions, y_prob=probablities)
    for metric in test_single_task_benchmark_multi_clf.metrics:
        assert metric in result.results.Metric.tolist()


def test_metrics_multitask_clf(tmpdir: str, test_multi_task_benchmark_clf: MultiTaskBenchmarkSpecification):
    train, test = test_multi_task_benchmark_clf.get_train_test_split()
    predictions = {
        target_col: np.random.randint(2, size=test.inputs.shape[0]) for target_col in train.target_cols
    }
    probabilities = {
        target_col: np.random.uniform(size=test.inputs.shape[0]) for target_col in train.target_cols
    }
    result = test_multi_task_benchmark_clf.evaluate(y_pred=predictions, y_prob=probabilities)
    assert isinstance(result.results, pd.DataFrame)
    assert set(result.results.columns) == {
        "Test set",
        "Target label",
        "Metric",
        "Score",
    }
    # check the targets and metrics
    assert result.results.shape[0] == len(test_multi_task_benchmark_clf.target_cols) * len(
        test_multi_task_benchmark_clf.metrics
    )
    for metric in test_multi_task_benchmark_clf.metrics:
        assert metric in result.results.Metric.tolist()


def test_metric_direction():
    for metric in Metric:
        assert metric.value.direction in ["min", "max", 1]


def test_absolute_average_fold_error():
    y_true = np.random.uniform(low=50, high=100, size=200)
    y_pred_1 = y_true + np.random.uniform(low=0, high=5, size=200)
    y_pred_2 = y_true + np.random.uniform(low=5, high=20, size=200)
    y_pred_3 = y_true - 10
    y_zero = np.zeros(shape=200)

    # Optimal value
    aafe_0 = Metric.absolute_average_fold_error(y_true=y_true, y_pred=y_true)
    assert aafe_0 == 1

    # small fold change
    aafe_1 = Metric.absolute_average_fold_error(y_true=y_true, y_pred=y_pred_1)
    assert aafe_1 > 1

    # larger fold change
    aafe_2 = Metric.absolute_average_fold_error(y_true=y_true, y_pred=y_pred_2)
    assert aafe_2 > aafe_1

    # undershoot
    aafe_3 = Metric.absolute_average_fold_error(y_true=y_true, y_pred=y_pred_3)
    assert aafe_3 < 1

    # y_true contains zeros
    with pytest.raises(ValueError):
        Metric.absolute_average_fold_error(y_true=y_zero, y_pred=y_pred_3)


def test_metric_y_types(
    tmpdir: str, test_single_task_benchmark_clf: SingleTaskBenchmarkSpecification, test_data: Dataset
):
    # here we use train split for testing purpose.
    _, test = test_single_task_benchmark_clf.get_train_test_split()
    predictions = np.random.randint(2, size=test.inputs.shape[0])
    probabilities = np.random.uniform(size=test.inputs.shape[0])
    test_y = test_data.loc[test.indices, "CLASS_expt"]

    # If y_pred is None and y_prob is None, an error is thrown.
    with pytest.raises(ValueError, match="Neither `y_pred` nor `y_prob` is specified."):
        test_single_task_benchmark_clf.evaluate()

    # If y_type == "y_pred" and y_pred is None, an error is thrown.
    with pytest.raises(ValueError, match="Metric.accuracy requires `y_pred` input"):
        test_single_task_benchmark_clf.metrics = [Metric.accuracy]
        test_single_task_benchmark_clf.evaluate(y_prob=probabilities)

    # If y_type != "y_pred" and y_prob is None, an error is thrown.
    with pytest.raises(ValueError, match="Metric.roc_auc requires `y_prob` input"):
        test_single_task_benchmark_clf.metrics = [Metric.roc_auc]
        test_single_task_benchmark_clf.evaluate(y_pred=predictions)

    # If y_type != "y_pred" and y_prob is None, an error is thrown.
    with pytest.raises(ValueError, match="Metric.pr_auc requires `y_prob` input"):
        test_single_task_benchmark_clf.metrics = [Metric.pr_auc]
        test_single_task_benchmark_clf.evaluate(y_pred=predictions)

    # If y_type != "y_pred" and y_pred is not None and y_prob is not None, it uses y_prob as expected!
    test_single_task_benchmark_clf.metrics = [Metric.roc_auc]
    result = test_single_task_benchmark_clf.evaluate(y_pred=predictions, y_prob=probabilities)
    assert result.results.Score.values[0] == Metric.roc_auc.fn(y_true=test_y, y_score=probabilities)

    # If y_type == "y_pred" and y_pred is not None and y_prob is not None, it uses y_pred as expected!
    test_single_task_benchmark_clf.metrics = [Metric.f1]
    result = test_single_task_benchmark_clf.evaluate(y_pred=predictions, y_prob=probabilities)
    assert result.results.Score.values[0] == Metric.f1.fn(y_true=test_y, y_pred=predictions)



import datamol as dm
import numpy as np
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from polaris.evaluate import BenchmarkResults


def test_single_task_benchmark_loop(test_single_task_benchmark):
    """Tests the integrated API for a single-task benchmark."""
    train, test = test_single_task_benchmark.get_train_test_split()

    model = RandomForestRegressor()
    smiles, y = train.as_array("xy")
    x = np.array([dm.to_fp(dm.to_mol(smi)) for smi in smiles])
    model.fit(X=x, y=y)

    x = np.array([dm.to_fp(dm.to_mol(smi)) for smi in test.inputs])
    y_pred = model.predict(x)

    scores = test_single_task_benchmark.evaluate(y_pred)
    assert isinstance(scores, BenchmarkResults)


def test_single_task_benchmark_loop_with_multiple_test_sets(test_single_task_benchmark_multiple_test_sets):
    """Tests the integrated API for a single-task benchmark with multiple test sets."""
    train, test = test_single_task_benchmark_multiple_test_sets.get_train_test_split()

    smiles, y = train.as_array("xy")

    x_train = np.array([dm.to_fp(dm.to_mol(smi)) for smi in smiles])

    model = RandomForestRegressor()
    model.fit(X=x_train, y=y)

    y_pred = {}
    for k, test_subset in test.items():
        x_test = np.array([dm.to_fp(dm.to_mol(smi)) for smi in test_subset.inputs])
        y_pred[k] = model.predict(x_test)

    scores = test_single_task_benchmark_multiple_test_sets.evaluate(y_pred)
    assert isinstance(scores, BenchmarkResults)


def test_single_task_benchmark_clf_loop_with_multiple_test_sets(
    test_single_task_benchmark_clf_multiple_test_sets,
):
    """Tests the integrated API for a single-task benchmark for classification probabilities with multiple test sets."""
    train, test = test_single_task_benchmark_clf_multiple_test_sets.get_train_test_split()

    smiles, y = train.as_array("xy")

    x_train = np.array([dm.to_fp(dm.to_mol(smi)) for smi in smiles])

    model = RandomForestClassifier()
    model.fit(X=x_train, y=y)

    y_prob = {}
    y_pred = {}
    for k, test_subset in test.items():
        x_test = np.array([dm.to_fp(dm.to_mol(smi)) for smi in test_subset.inputs])
        y_prob[k] = model.predict_proba(x_test)[:, :1]  # for binary classification
        y_pred[k] = model.predict(x_test)

    scores = test_single_task_benchmark_clf_multiple_test_sets.evaluate(y_prob=y_prob, y_pred=y_pred)
    assert isinstance(scores, BenchmarkResults)


def test_multi_task_benchmark_loop(test_multi_task_benchmark):
    """Tests the integrated API for a multi-task benchmark."""
    train, test = test_multi_task_benchmark.get_train_test_split()

    smiles, multi_y = train.as_array("xy")
    x_train = np.array([dm.to_fp(dm.to_mol(smi)) for smi in smiles])
    x_test = np.array([dm.to_fp(dm.to_mol(smi)) for smi in test.inputs])

    y_pred = {}
    for k, y in multi_y.items():
        model = RandomForestRegressor()

        mask = ~np.isnan(y)
        model.fit(X=x_train[mask], y=y[mask])
        y_pred[k] = model.predict(x_test)

    scores = test_multi_task_benchmark.evaluate(y_pred)
    assert isinstance(scores, BenchmarkResults)



import mmap

import zarr


class MemoryMappedDirectoryStore(zarr.DirectoryStore):
    """
    A Zarr Store to open chunks as memory-mapped files.
    See also [this Github issue](https://github.com/zarr-developers/zarr-python/issues/1245).

    Memory mapping leverages low-level OS functionality to reduce the time it takes
    to read the content of a file by directly mapping to memory.
    """

    def _fromfile(self, fn):
        with open(fn, "rb") as fh:
            return memoryview(mmap.mmap(fh.fileno(), 0, prot=mmap.PROT_READ))



import datamol as dm
import pandas as pd
import pytest

from polaris.dataset import DatasetFactory, create_dataset_from_file
from polaris.dataset.converters import SDFConverter, ZarrConverter


def _check_dataset(dataset, ground_truth, mol_props_as_col):
    assert len(dataset) == 1

    mol = dataset.get_data(row=0, col="molecule")

    assert isinstance(mol, dm.Mol)

    if mol_props_as_col:
        assert not mol.HasProp("my_property")
        v = dataset.get_data(row=0, col="my_property")
        assert v == ground_truth.GetProp("my_property")

    else:
        assert mol.HasProp("my_property")
        assert mol.GetProp("my_property") == ground_truth.GetProp("my_property")
        assert "my_property" not in dataset.columns


def test_sdf_zarr_conversion(sdf_file, caffeine, tmpdir):
    """Test conversion between SDF and Zarr with utility function"""
    dataset = create_dataset_from_file(sdf_file, tmpdir.join("archive.zarr"))
    _check_dataset(dataset, caffeine, True)


@pytest.mark.parametrize("mol_props_as_col", [True, False])
def test_factory_sdf_with_prop_as_col(sdf_file, caffeine, tmpdir, mol_props_as_col):
    """Test conversion between SDF and Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("archive.zarr"))

    converter = SDFConverter(mol_prop_as_cols=mol_props_as_col)
    factory.register_converter("sdf", converter)

    factory.add_from_file(sdf_file)
    dataset = factory.build()

    _check_dataset(dataset, caffeine, mol_props_as_col)


def test_zarr_to_zarr_conversion(zarr_archive, tmpdir):
    """Test conversion between Zarr and Zarr with utility function"""
    dataset = create_dataset_from_file(zarr_archive, tmpdir.join("archive.zarr"))
    assert len(dataset) == 100
    assert len(dataset.columns) == 2
    assert all(c in dataset.columns for c in ["A", "B"])
    assert all(dataset.annotations[c].is_pointer for c in ["A", "B"])
    assert dataset.get_data(row=0, col="A").shape == (2048,)


def test_zarr_with_factory_pattern(zarr_archive, tmpdir):
    """Test conversion between Zarr and Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("archive.zarr"))
    converter = ZarrConverter()
    factory.register_converter("zarr", converter)
    factory.add_from_file(zarr_archive)

    factory.add_column(pd.Series([1, 2, 3, 4] * 25, name="C"))

    df = pd.DataFrame({"C": [1, 2, 3, 4], "D": ["W", "X", "Y", "Z"]})
    factory.add_columns(df, merge_on="C")

    dataset = factory.build()
    assert len(dataset) == 100
    assert len(dataset.columns) == 4
    assert all(c in dataset.columns for c in ["A", "B", "C", "D"])
    assert dataset.table["C"].apply({1: "W", 2: "X", 3: "Y", 4: "Z"}.get).equals(dataset.table["D"])



import json
import uuid
from hashlib import md5
from pathlib import Path
from typing import Dict, List, MutableMapping, Optional, Tuple, Union

import fsspec
import numpy as np
import pandas as pd
import zarr
from datamol.utils import fs as dmfs
from loguru import logger
from pydantic import (
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)

from polaris._artifact import BaseArtifactModel
from polaris.mixins import ChecksumMixin
from polaris.dataset._adapters import Adapter
from polaris.dataset._column import ColumnAnnotation
from polaris.dataset.zarr import MemoryMappedDirectoryStore, ZarrFileChecksum, compute_zarr_checksum
from polaris.hub.polarisfs import PolarisFileSystem
from polaris.utils.constants import DEFAULT_CACHE_DIR
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidDatasetError
from polaris.utils.types import (
    AccessType,
    HttpUrlString,
    HubOwner,
    SupportedLicenseType,
    ZarrConflictResolution,
)

# Constants
_SUPPORTED_TABLE_EXTENSIONS = ["parquet"]
_CACHE_SUBDIR = "datasets"
_INDEX_SEP = "#"


class Dataset(BaseArtifactModel, ChecksumMixin):
    """Basic data-model for a Polaris dataset, implemented as a [Pydantic](https://docs.pydantic.dev/latest/) model.

    At its core, a dataset in Polaris is a tabular data structure that stores data-points in a row-wise manner.
    A Dataset can have multiple modalities or targets, can be sparse and can be part of one or multiple
     [`BenchmarkSpecification`][polaris.benchmark.BenchmarkSpecification] objects.

    Info: Pointer columns
        Whereas a `Dataset` contains all information required to construct a dataset, it is not ready yet.
        For complex data, such as images, we support storing the content in external blobs of data.
        In that case, the table contains _pointers_ to these blobs that are dynamically loaded when needed.

    Attributes:
        table: The core data-structure, storing data-points in a row-wise manner. Can be specified as either a
            path to a `.parquet` file or a `pandas.DataFrame`.
        default_adapters: The adapters that the Dataset recommends to use by default to change the format of the data
            for specific columns.
        zarr_root_path: The data for any pointer column should be saved in the Zarr archive this path points to.
        readme: Markdown text that can be used to provide a formatted description of the dataset.
            If using the Polaris Hub, it is worth noting that this field is more easily edited through the Hub UI
            as it provides a rich text editor for writing markdown.
        annotations: Each column _can be_ annotated with a [`ColumnAnnotation`][polaris.dataset.ColumnAnnotation] object.
            Importantly, this is used to annotate whether a column is a pointer column.
        source: The data source, e.g. a DOI, Github repo or URI.
        license: The dataset license. Polaris only supports some Creative Commons licenses. See [`SupportedLicenseType`][polaris.utils.types.SupportedLicenseType] for accepted ID values.
        curation_reference: A reference to the curation process, e.g. a DOI, Github repo or URI.
    For additional meta-data attributes, see the [`BaseArtifactModel`][polaris._artifact.BaseArtifactModel] class.

    Raises:
        InvalidDatasetError: If the dataset does not conform to the Pydantic data-model specification.
    """

    # Public attributes
    # Data
    table: Union[pd.DataFrame, str]
    default_adapters: Dict[str, Adapter] = Field(default_factory=dict)
    zarr_root_path: Optional[str] = None

    # Additional meta-data
    readme: str = ""
    annotations: Dict[str, ColumnAnnotation] = Field(default_factory=dict)
    source: Optional[HttpUrlString] = None
    license: Optional[SupportedLicenseType] = None
    curation_reference: Optional[HttpUrlString] = None

    # Config
    cache_dir: Optional[Path] = None  # Where to cache the data to if cache() is called.

    # Private attributes
    _zarr_root: Optional[zarr.Group] = PrivateAttr(None)
    _zarr_data: Optional[MutableMapping[str, np.ndarray]] = PrivateAttr(None)
    _md5sum: Optional[str] = PrivateAttr(None)
    _zarr_md5sum_manifest: List[ZarrFileChecksum] = PrivateAttr(default_factory=list)
    _client = PrivateAttr(None)  # Optional[PolarisHubClient]
    _warn_about_remote_zarr: bool = PrivateAttr(True)

    @field_validator("table")
    def _validate_table(cls, v):
        """
        If the table is not a dataframe yet, assume it's a path and try load it.
        We also make sure that the pandas index is contiguous and starts at 0, and
        that all columns are named and unique.
        """
        # Load from path if not a dataframe
        if not isinstance(v, pd.DataFrame):
            if not dmfs.is_file(v) or dmfs.get_extension(v) not in _SUPPORTED_TABLE_EXTENSIONS:
                raise InvalidDatasetError(f"{v} is not a valid DataFrame or .parquet path.")
            v = pd.read_parquet(v)
        # Check if there are any duplicate columns
        if any(v.columns.duplicated()):
            raise InvalidDatasetError("The table contains duplicate columns")
        # Check if there are any unnamed columns
        if not all(isinstance(c, str) for c in v.columns):
            raise InvalidDatasetError("The table contains unnamed columns")
        return v

    @model_validator(mode="after")
    @classmethod
    def _validate_model(cls, m: "Dataset"):
        """Verifies some dependencies between properties"""

        # Verify that all annotations are for columns that exist
        if any(k not in m.table.columns for k in m.annotations):
            raise InvalidDatasetError("There are annotations for columns that do not exist")

        # Verify that all adapters are for columns that exist
        if any(k not in m.table.columns for k in m.default_adapters.keys()):
            raise InvalidDatasetError("There are default adapters for columns that do not exist")

        has_pointers = any(anno.is_pointer for anno in m.annotations.values())
        if has_pointers and m.zarr_root_path is None:
            raise InvalidDatasetError("A zarr_root_path needs to be specified when there are pointer columns")
        if not has_pointers and m.zarr_root_path is not None:
            raise InvalidDatasetError(
                "The zarr_root_path should only be specified when there are pointer columns"
            )

        # Set a default for missing annotations and convert strings to Modality
        for c in m.table.columns:
            if c not in m.annotations:
                m.annotations[c] = ColumnAnnotation()
            m.annotations[c].dtype = m.table[c].dtype

        # Set the default cache dir if none and make sure it exists
        if m.cache_dir is None:
            dataset_id = m._md5sum if m.has_md5sum else str(uuid.uuid4())
            m.cache_dir = Path(DEFAULT_CACHE_DIR) / _CACHE_SUBDIR / dataset_id

        m.cache_dir.mkdir(parents=True, exist_ok=True)
        return m

    @field_validator("default_adapters", mode="before")
    def _validate_adapters(cls, value):
        """Validate the adapters"""
        return {k: Adapter[v] if isinstance(v, str) else v for k, v in value.items()}

    @field_serializer("default_adapters")
    def _serialize_adapters(self, value: List[Adapter]):
        """Serializes the adapters"""
        return {k: v.name for k, v in value.items()}

    @field_serializer("cache_dir")
    def _serialize_cache_dir(value):
        """Serialize the cacha_dir"""
        return str(value)

    def _compute_checksum(self):
        """Computes a hash of the dataset.

        This is meant to uniquely identify the dataset and can be used to verify the version.

        1. Is not sensitive to the ordering of the columns or rows in the table.
        2. Purposefully does not include the meta-data (source, description, name, annotations).
        3. Includes a hash for the Zarr archive.
        """
        hash_fn = md5()

        # Sort the columns s.t. the checksum is not sensitive to the column-ordering
        df = self.table.copy(deep=True)
        df = df[sorted(df.columns.tolist())]

        # Use the sum of the row-wise hashes s.t. the hash is insensitive to the row-ordering
        table_hash = pd.util.hash_pandas_object(df, index=False).sum()
        hash_fn.update(table_hash)

        # If the Zarr archive exists, we hash its contents too.
        if self.uses_zarr:
            zarr_hash, self._zarr_md5sum_manifest = compute_zarr_checksum(self.zarr_root_path)
            hash_fn.update(zarr_hash.encode())

        checksum = hash_fn.hexdigest()
        return checksum

    @computed_field
    @property
    def zarr_md5sum_manifest(self) -> List[ZarrFileChecksum]:
        """
        The Zarr Checksum manifest stores the checksums of all files in a Zarr archive.
        If the dataset doesn't use Zarr, this will simply return an empty list.
        """
        if len(self._zarr_md5sum_manifest) == 0 and not self.has_md5sum:
            # The manifest is set as an instance variable
            # as a side-effect of the compute_checksum method
            self.md5sum = self._compute_checksum()
        return self._zarr_md5sum_manifest

    @property
    def client(self):
        """The Polaris Hub client used to interact with the Polaris Hub."""

        # Import it here to prevent circular imports
        from polaris.hub.client import PolarisHubClient

        if self._client is None:
            self._client = PolarisHubClient()
        return self._client

    @property
    def uses_zarr(self) -> bool:
        """Whether any of the data in this dataset is stored in a Zarr Archive."""
        return self.zarr_root_path is not None

    @property
    def zarr_data(self):
        """Get the Zarr data.

        This is different from the Zarr Root, because to optimize the efficiency of
        data loading, a user can choose to load the data into memory as a numpy array

        Note: General purpose dataloader.
            The goal with Polaris is to provide general purpose datasets that serve as good
            options for a _wide variety of use cases_. This also implies you should be able to
            optimize things further for a specific use case if needed.
        """
        if self._zarr_data is not None:
            return self._zarr_data
        return self.zarr_root

    @property
    def zarr_root(self):
        """Get the zarr Group object corresponding to the root.

        Opens the zarr archive in read-write mode if it is not already open.

        Note: Different to `zarr_data`
            The `zarr_data` attribute references either to the Zarr archive or to a in-memory copy of the data.
            See also [`Dataset.load_to_memory`][polaris.dataset.Dataset.load_to_memory].
        """

        if self._zarr_root is not None:
            return self._zarr_root

        if self.zarr_root_path is None:
            return None

        # We open the archive in read-only mode if it is saved on the Hub
        saved_on_hub = PolarisFileSystem.is_polarisfs_path(self.zarr_root_path)

        if self._warn_about_remote_zarr:
            saved_remote = saved_on_hub or not Path(self.zarr_root_path).exists()

            if saved_remote:
                logger.warning(
                    f"You're loading data from a remote location. "
                    f"To speed up this process, consider caching the dataset first "
                    f"using {self.__class__.__name__}.cache()"
                )
                self._warn_about_remote_zarr = False

        try:
            if saved_on_hub:
                self._zarr_root = self.client.open_zarr_file(self.owner, self.name, self.zarr_root_path, "r+")
            else:
                # We use memory mapping by default because our experiments show that it's consistently faster
                store = MemoryMappedDirectoryStore(self.zarr_root_path)
                self._zarr_root = zarr.open_consolidated(store, mode="r+")
        except KeyError as error:
            raise InvalidDatasetError(
                "A Zarr archive associated with a Polaris dataset has to be consolidated."
            ) from error
        return self._zarr_root

    @computed_field
    @property
    def n_rows(self) -> int:
        """The number of rows in the dataset."""
        return len(self.rows)

    @computed_field
    @property
    def n_columns(self) -> int:
        """The number of columns in the dataset."""
        return len(self.columns)

    @property
    def rows(self) -> list:
        """Return all row indices for the dataset"""
        return self.table.index.tolist()

    @property
    def columns(self) -> list:
        """Return all columns for the dataset"""
        return self.table.columns.tolist()

    def load_to_memory(self):
        """Pre-load the entire dataset into memory.

        Warning: Make sure the uncompressed dataset fits in-memory.
            This method will load the **uncompressed** dataset into memory. Make
            sure you actually have enough memory to store the dataset.
        """
        data = self.zarr_data

        if not isinstance(data, zarr.Group):
            raise TypeError(
                "The dataset zarr_root is not a valid Zarr archive. "
                "Did you call Dataset.load_to_memory() twice?"
            )

        # NOTE (cwognum): If the dataset fits in memory, we are best of using plain NumPy arrays.
        # Even if we disable chunking and compression in Zarr.
        # For more information, see https://github.com/zarr-developers/zarr-python/issues/1395
        self._zarr_data = {k: data[k][:] for k in data.array_keys()}

    def get_data(self, row: str | int, col: str, adapters: Optional[List[Adapter]] = None) -> np.ndarray:
        """Since the dataset might contain pointers to external files, data retrieval is more complicated
        than just indexing the `table` attribute. This method provides an end-point for seamlessly
        accessing the underlying data.

        Args:
            row: The row index in the `Dataset.table` attribute
            col: The column index in the `Dataset.table` attribute
            adapters: The adapters to apply to the data before returning it.
                If None, will use the default adapters specified for the dataset.

        Returns:
            A numpy array with the data at the specified indices. If the column is a pointer column,
                the content of the referenced file is loaded to memory.
        """

        # Fetch adapters for dataset and a given column
        adapters = adapters or self.default_adapters
        adapter = adapters.get(col)

        # If not a pointer, return it here. Apply adapter if specified.
        value = self.table.at[row, col]
        if not self.annotations[col].is_pointer:
            if adapter is not None:
                return adapter(value)
            return value

        # Load the data from the Zarr archive
        path, index = self._split_index_from_path(value)
        arr = self.zarr_data[path][index]

        # Change to tuple if a slice
        if isinstance(index, slice):
            arr = tuple(arr)

        # Adapt the input to the specified format
        if adapter is not None:
            arr = adapter(arr)

        return arr

    def upload_to_hub(
        self, access: Optional[AccessType] = "private", owner: Optional[Union[HubOwner, str]] = None
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_dataset`][polaris.hub.client.PolarisHubClient.upload_dataset] method.
        """
        self.client.upload_dataset(self, access=access, owner=owner)

    @classmethod
    def from_json(cls, path: str):
        """Loads a benchmark from a JSON file.
        Overrides the method from the base class to remove the caching dir from the file to load from,
        as that should be user dependent.

        Args:
            path: Loads a benchmark specification from a JSON file.
        """
        with fsspec.open(path, "r") as f:
            data = json.load(f)
        data.pop("cache_dir", None)
        return cls.model_validate(data)

    def to_json(
        self,
        destination: str,
        if_exists: ZarrConflictResolution = "replace",
    ) -> str:
        """
        Save the dataset to a destination directory as a JSON file.

        Warning: Multiple files
            Perhaps unintuitive, this method creates multiple files.

            1. `/path/to/destination/dataset.json`: This file can be loaded with
                [`Dataset.from_json`][polaris.dataset.Dataset.from_json].
            2. `/path/to/destination/table.parquet`: The `Dataset.table` attribute is saved here.
            3. _(Optional)_ `/path/to/destination/data/*`: Any additional blobs of data referenced by the
                    pointer columns will be stored here.

        Args:
            destination: The _directory_ to save the associated data to.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.

        Returns:
            The path to the JSON file.
        """
        dmfs.mkdir(destination, exist_ok=True)
        table_path = dmfs.join(destination, "table.parquet")
        dataset_path = dmfs.join(destination, "dataset.json")
        new_zarr_root_path = dmfs.join(destination, "data.zarr")

        # Lu: Avoid serilizing and sending None to hub app.
        serialized = self.model_dump(exclude={"cache_dir"}, exclude_none=True)
        serialized["table"] = table_path

        # Copy over Zarr data to the destination
        if self.uses_zarr:
            self._warn_about_remote_zarr = False

            logger.info(f"Copying Zarr archive to {new_zarr_root_path}. This may take a while.")

            dest = zarr.open(new_zarr_root_path, "w")

            zarr.copy_store(
                source=self.zarr_root.store.store,
                dest=dest.store,
                log=logger.debug,
                if_exists=if_exists,
            )

        self.table.to_parquet(table_path)
        with fsspec.open(dataset_path, "w") as f:
            json.dump(serialized, f)

        return dataset_path

    def cache(self, cache_dir: Optional[str] = None, verify_checksum: bool = True) -> str:
        """Caches the dataset by downloading all additional data for pointer columns to a local directory.

        Args:
            cache_dir: The directory to cache the data to. If not provided,
                this will fall back to the `Dataset.cache_dir` attribute
            verify_checksum: Whether to verify the checksum of the dataset after caching.

        Returns:
            The path to the cache directory.
        """

        if cache_dir is not None:
            self.cache_dir = cache_dir

        self.to_json(self.cache_dir)

        if self.uses_zarr:
            self.zarr_root_path = dmfs.join(self.cache_dir, "data.zarr")
            self._zarr_root = None

        if verify_checksum:
            self.verify_checksum()

        return self.cache_dir

    def size(self):
        return self.rows, self.n_columns

    def _split_index_from_path(self, path: str) -> Tuple[str, Optional[int]]:
        """
        Paths can have an additional index appended to them.
        This extracts that index from the path.
        """
        index = None
        if _INDEX_SEP in path:
            path, index = path.split(_INDEX_SEP)
            index = index.split(":")

            if len(index) == 1:
                index = int(index[0])
            elif len(index) == 2:
                index = slice(int(index[0]), int(index[1]))
            else:
                raise ValueError(f"Invalid index format: {index}")
        return path, index

    def __getitem__(self, item):
        """Allows for indexing the dataset directly"""
        ret = self.table.loc[item]
        if isinstance(ret, pd.Series):
            # Load the data from the pointer columns

            if ret.name in self.table.columns:
                # Returning a column, the indices are rows
                if self.annotations[ret.name].is_pointer:
                    ret = np.array([self.get_data(k, ret.name) for k in ret.index])

            elif len(ret) == self.n_rows:
                # Returning a row, the indices are columns
                ret = {
                    k: self.get_data(k, ret.name) if self.annotations[ret.name].is_pointer else ret[k]
                    for k in ret.index
                }

        # Returning a dataframe
        if isinstance(ret, pd.DataFrame):
            for c in ret.columns:
                if self.annotations[c].is_pointer:
                    ret[c] = [self.get_data(item, c) for item in ret.index]
            return ret

        return ret

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump(exclude={"table", "zarr_md5sum_manifest"})
        return repr_dict

    def _repr_html_(self):
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self._repr_dict_())

    def __len__(self):
        return self.n_rows

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        """Whether two datasets are equal is solely determined by the checksum"""
        if not isinstance(other, Dataset):
            return False
        return self.md5sum == other.md5sum

    def __del__(self):
        """Close the connection of the client"""
        if self._client is not None:
            self._client.close()



import os
from typing import Dict, Optional

import datamol as dm
import pandas as pd
import zarr
from loguru import logger

from polaris.dataset import ColumnAnnotation, Dataset
from polaris.dataset._adapters import Adapter
from polaris.dataset.converters import Converter, SDFConverter, ZarrConverter


def create_dataset_from_file(path: str, zarr_root_path: Optional[str] = None) -> Dataset:
    """
    This function is a convenience function to create a dataset from a file.

    It sets up the dataset factory with sensible defaults for the converters.
    For creating more complicated datasets, please use the `DatasetFactory` directly.
    """
    factory = DatasetFactory(zarr_root_path=zarr_root_path)
    factory.register_converter("sdf", SDFConverter())
    factory.register_converter("zarr", ZarrConverter())

    factory.add_from_file(path)
    return factory.build()


class DatasetFactory:
    """
    The `DatasetFactory` makes it easier to create complex datasets.

    It is based on the the factory design pattern and allows a user to specify specific handlers
    (i.e. [`Converter`][polaris.dataset.converters._base.Converter] objects) for different file types.
    These converters are used to convert commonly used file types in drug discovery
    to something that can be used within Polaris while losing as little information as possible.

    In addition, it contains utility method to incrementally build out a dataset from different sources.

    Tip: Try quickly converting one of your datasets
        The `DatasetFactory` is designed to give you full control.
        If your dataset is saved in a single file and you don't need anything fancy, you can try use
        [`create_dataset_from_file`][polaris.dataset.create_dataset_from_file] instead.

        ```py
        from polaris.dataset import create_dataset_from_file
        dataset = create_dataset_from_file("path/to/my_dataset.sdf")
        ```

    Question: How to make adding meta-data easier?
        The `DatasetFactory` is designed to more easily pull together data from different sources.
        However, adding meta-data remains a laborous process. How could we make this simpler through
        the Python API?
    """

    def __init__(
        self, zarr_root_path: Optional[str] = None, converters: Optional[Dict[str, Converter]] = None
    ) -> None:
        """
        Create a new factory object.

        Args:
            zarr_root_path: The root path of the zarr hierarchy. If you want to use pointer columns,
                this arguments needs to be passed.
            converters: The converters to use for specific file types.
                You can also register them later with register_converter().
        """

        if converters is None:
            converters = {}

        self._converters: Dict[str, Converter] = converters
        self.reset(zarr_root_path=zarr_root_path)

    @property
    def zarr_root_path(self) -> zarr.Group:
        """
        The root of the zarr archive for the Dataset that is being built.
        All data for a single dataset is expected to be stored in the same Zarr archive.
        """
        if self._zarr_root_path is None:
            raise ValueError("You need to pass `zarr_root_path` to the factory to use pointer columns")
        return self._zarr_root_path

    @property
    def zarr_root(self) -> zarr.Group:
        """
        The root of the zarr archive for the Dataset that is being built.
        All data for a single dataset is expected to be stored in the same Zarr archive.
        """
        if self._zarr_root is None:
            # NOTE (cwognum): The DirectoryStore is the default store when calling zarr.open
            #   I nevertheless explicitly set it here to make it clear that this is a design decision.
            #   We could consider using different stores, such as the NestedDirectoryStore.
            store = zarr.DirectoryStore(self.zarr_root_path)
            self._zarr_root = zarr.open(store, "w")
            if not isinstance(self._zarr_root, zarr.Group):
                raise ValueError("The root of the zarr hierarchy should be a group")
        return self._zarr_root

    def register_converter(self, ext: str, converter: Converter):
        """
        Registers a new converter for a specific file type.

        Args:
            ext: The file extension for which the converter should be used.
                There can only be a single converter per file extension.
            converter: The handler for the file type. This should convert
                the file to a Polaris-compatible format.
        """
        if ext in self._converters:
            logger.info(f"You are overwriting the converter for the {ext} extension.")
        self._converters[ext] = converter

    def add_column(
        self,
        column: pd.Series,
        annotation: Optional[ColumnAnnotation] = None,
        adapters: Optional[Adapter] = None,
    ):
        """
        Add a single column to the DataFrame

        We require:

        1. The name attribute of the column to be set.
        2. The name attribute of the column to be unique.
        3. If the column is a pointer column, the `zarr_root_path` needs to be set.
        4. The length of the column to match the length of the alredy constructed table.

        Args:
            column: The column to add to the dataset.
            annotation: The annotation for the column. If None, a default annotation will be used.
        """

        # Verify the column can be added
        if column.name is None:
            raise ValueError("You need to specify a column name")
        if column.name in self._table.columns:
            raise ValueError(f"Column name '{column.name}' already exists in the table")
        if not self._table.empty and len(column) != len(self._table):
            raise ValueError("The length of the column does not match the length of the table")

        if annotation is not None and annotation.is_pointer:
            if self.zarr_root is None:
                raise ValueError("You need to pass `zarr_root_path` to the factory to use pointer columns")

        # Actually add the column
        self._table[column.name] = column

        if annotation is None:
            annotation = ColumnAnnotation()
        self._annotations[column.name] = annotation

        if adapters is not None:
            self._adapters[column.name] = adapters

    def add_columns(
        self,
        df: pd.DataFrame,
        annotations: Optional[Dict[str, ColumnAnnotation]] = None,
        adapters: Optional[Dict[str, Adapter]] = None,
        merge_on: Optional[str] = None,
    ):
        """
        Add multiple columns to the dataset based on another dataframe.

        To have more control over how the two dataframes are combined, you can
        specify a column to merge on. This will always do an **outer** join.

        If not specifying a key to merge on, the columns will simply be added to the dataset
        that has been built so far without any reordering. They are therefore expected to meet all
        the same expectations as for [`add_column`][polaris.dataset.DatasetFactory.add_column].

        Args:
            df: A Pandas DataFrame with the columns that we want to add to the dataset.
            annotations: The annotations for the columns. If None, default annotations will be used.
            merge_on: The column to merge on, if any.
        """

        if merge_on is not None:
            df = self._table.merge(df, on=merge_on, how="outer")

        if annotations is None:
            annotations = {}
        annotations = {**self._annotations, **annotations}

        if adapters is None:
            adapters = {}
        adapters = {**self._adapters, **adapters}

        if merge_on is not None:
            self.reset(self._zarr_root_path)

        for name, series in df.items():
            annotation = annotations.get(name)
            adapter = adapters.get(name)
            self.add_column(series, annotation, adapter)

    def add_from_file(self, path: str):
        """
        Uses the registered converters to parse the data from a specific file and add it to the dataset.
        If no converter is found for the file extension, it raises an error.

        Args:
            path: The path to the file that should be parsed.
        """
        ext = dm.fs.get_extension(path)
        converter = self._converters.get(ext)
        if converter is None:
            raise ValueError(f"No converter found for extension {ext}")

        table, annotations, adapters = converter.convert(path, self)
        self.add_columns(table, annotations, adapters)

    def build(self) -> Dataset:
        """Returns a Dataset based on the current state of the factory."""
        zarr.consolidate_metadata(self.zarr_root.store)
        return Dataset(
            table=self._table,
            annotations=self._annotations,
            default_adapters=self._adapters,
            zarr_root_path=self.zarr_root_path,
        )

    def reset(self, zarr_root_path: Optional[str] = None):
        """
        Resets the factory to its initial state to start building the next dataset from scratch.
        Note that this will not reset the registered converters.

        Args:
            zarr_root_path: The root path of the zarr hierarchy. If you want to use pointer columns
                for your next dataset, this arguments needs to be passed.
        """

        if zarr_root_path is not None:
            zarr_root_path = os.path.abspath(zarr_root_path).rstrip("/")

        self._zarr_root = None
        self._zarr_root_path = zarr_root_path
        self._table = pd.DataFrame()
        self._annotations = {}
        self._adapters = {}



import warnings

import pytest
from pydantic import BaseModel, ValidationError

import polaris as po
from polaris._artifact import BaseArtifactModel
from polaris.utils.types import HttpUrlString, HubOwner


def test_slug_string_type():
    """
    Verifies that the slug is validated correctly.
    Fails if:
    - Is too short (<4 characters)
    - Is too long (>64 characters)
    - Contains something other than lowercase letters, numbers, and hyphens.
    """
    for name in [
        "",
        "x",
        "xx",
        "xxx",
        "x" * 65,
        "invalid@",
        "invalid!",
        "InvalidName1",
        "invalid_name",
    ]:
        with pytest.raises(ValidationError):
            HubOwner(slug=name)

    for name in ["valid", "valid-name-1", "x" * 64, "x" * 4]:
        HubOwner(slug=name)


def test_slug_compatible_string_type():
    """Verifies that the artifact name is validated correctly."""

    # Fails if:
    # - Is too short (<4 characters)
    # - Is too long (>64 characters)
    # - Contains non-alphanumeric characters
    for name in ["", "x", "xx", "xxx", "x" * 65, "invalid@", "invalid!"]:
        with pytest.raises(ValidationError):
            BaseArtifactModel(name=name)

    # Does not fail
    for name in [
        "valid",
        "valid-name",
        "valid_name",
        "ValidName1",
        "Valid_",
        "Valid-",
        "x" * 64,
        "x" * 4,
    ]:
        BaseArtifactModel(name=name)


def test_version():
    with pytest.raises(ValidationError):
        BaseArtifactModel(polaris_version="invalid")
    assert BaseArtifactModel().polaris_version == po.__version__
    assert BaseArtifactModel(polaris_version="0.1.2")


def test_http_url_string():
    """Verifies that a string validated correctly as a URL."""

    class _TestModel(BaseModel):
        url: HttpUrlString

    m = _TestModel(url="https://example.com")
    assert isinstance(m.url, str)

    m = _TestModel(url="http://example.com")
    assert isinstance(m.url, str)

    m = _TestModel(url="http://example.io")
    assert isinstance(m.url, str)

    with warnings.catch_warnings():
        # Crash if any warnings are raised
        warnings.simplefilter("error")
        m.model_dump()

    with pytest.raises(ValidationError):
        _TestModel(url="invalid")
    with pytest.raises(ValidationError):
        _TestModel(url="ftp://invalid.com")



from time import perf_counter

import numpy as np
import pandas as pd
import pytest
import zarr
from datamol.utils import fs

from polaris.dataset import Dataset, Subset, create_dataset_from_file
from polaris.loader import load_dataset


@pytest.mark.parametrize("with_caching", [True, False])
@pytest.mark.parametrize("with_slice", [True, False])
def test_load_data(tmp_path, with_slice, with_caching):
    """Test accessing the data, specifically whether pointer columns are properly handled."""

    # Dummy data (could e.g. be a 3D structure or Image)
    arr = np.random.random((100, 100))

    tmpdir = str(tmp_path)
    zarr_path = fs.join(tmpdir, "data.zarr")

    root = zarr.open(zarr_path, "w")
    root.array("A", data=arr)
    zarr.consolidate_metadata(root.store)

    path = "A#0:5" if with_slice else "A#0"
    table = pd.DataFrame({"A": [path]}, index=[0])
    dataset = Dataset(table=table, annotations={"A": {"is_pointer": True}}, zarr_root_path=zarr_path)

    if with_caching:
        dataset.cache(fs.join(tmpdir, "cache"))

    data = dataset.get_data(row=0, col="A")

    if with_slice:
        assert isinstance(data, tuple)
        assert len(data) == 5

        for i, d in enumerate(data):
            assert (d == arr[i]).all()

    else:
        data = dataset.get_data(row=0, col="A")
        assert (data == arr[0]).all()


def test_dataset_checksum(test_dataset):
    """Test whether the checksum is a good indicator of whether the dataset has changed in a meaningful way."""

    # Make sure the `md5sum` is part of the model dump even if not initiated yet.
    # This is important for uploads to the Hub.
    assert test_dataset._md5sum is None
    assert "md5sum" in test_dataset.model_dump()

    # Without any changes, same hash
    kwargs = test_dataset.model_dump()
    assert Dataset(**kwargs) == test_dataset

    # With unimportant changes, same hash
    kwargs["name"] = "changed"
    kwargs["description"] = "changed"
    kwargs["source"] = "https://changed.com"
    assert Dataset(**kwargs) == test_dataset

    # Check sensitivity to the row and column ordering
    kwargs["table"] = kwargs["table"].iloc[::-1]
    kwargs["table"] = kwargs["table"][kwargs["table"].columns[::-1]]
    assert Dataset(**kwargs) == test_dataset

    # Without any changes, but different hash
    dataset = Dataset(**kwargs)
    dataset._md5sum = "invalid"
    assert dataset != test_dataset

    # With changes, but same hash
    kwargs["md5sum"] = test_dataset.md5sum
    kwargs["table"] = kwargs["table"].iloc[:-1]
    assert Dataset(**kwargs) != test_dataset


def test_dataset_from_zarr(zarr_archive, tmpdir):
    """Test whether loading works when the zarr archive contains a single array or multiple arrays."""
    archive = zarr_archive
    dataset = create_dataset_from_file(archive, tmpdir.join("data"))

    assert len(dataset.table) == 100
    for i in range(100):
        assert dataset.get_data(row=i, col="A").shape == (2048,)
        assert dataset.get_data(row=i, col="B").shape == (2048,)


def test_dataset_from_json(test_dataset, tmpdir):
    """Test whether the dataset can be saved and loaded from json."""
    test_dataset.to_json(str(tmpdir))

    path = fs.join(str(tmpdir), "dataset.json")

    new_dataset = Dataset.from_json(path)
    assert test_dataset == new_dataset

    new_dataset = load_dataset(path)
    assert test_dataset == new_dataset


def test_dataset_from_zarr_to_json_and_back(zarr_archive, tmpdir):
    """
    Test whether a dataset with pointer columns, instantiated from a zarr archive,
    can be saved to and loaded from json.
    """

    json_dir = tmpdir.join("json")
    zarr_dir = tmpdir.join("zarr")

    archive = zarr_archive
    dataset = create_dataset_from_file(archive, zarr_dir)
    path = dataset.to_json(json_dir)

    new_dataset = Dataset.from_json(path)
    assert dataset == new_dataset

    new_dataset = load_dataset(path)
    assert dataset == new_dataset


def test_dataset_caching(zarr_archive, tmpdir):
    """Test whether the dataset remains the same after caching."""

    original_dataset = create_dataset_from_file(zarr_archive, tmpdir.join("original1"))
    cached_dataset = create_dataset_from_file(zarr_archive, tmpdir.join("original2"))
    assert original_dataset == cached_dataset

    cache_dir = cached_dataset.cache(tmpdir.join("cached").strpath)
    assert cached_dataset.zarr_root_path.startswith(cache_dir)

    assert cached_dataset == original_dataset


def test_dataset_index():
    """Small test to check whether the dataset resets its index."""
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]}, index=["X", "Y", "Z"])
    dataset = Dataset(table=df)
    subset = Subset(dataset=dataset, indices=[1], input_cols=["A"], target_cols=["B"])
    assert next(iter(subset)) == (np.array([2]), np.array([5]))


def test_dataset_in_memory_optimization(zarr_archive, tmpdir):
    """Check if optimization makes a default Zarr archive faster."""
    dataset = create_dataset_from_file(zarr_archive, tmpdir.join("dataset"))
    subset = Subset(dataset=dataset, indices=range(100), input_cols=["A"], target_cols=["B"])

    t1 = perf_counter()
    for x in subset:
        pass
    d1 = perf_counter() - t1

    dataset.load_to_memory()

    t2 = perf_counter()
    for x in subset:
        pass
    d2 = perf_counter() - t2

    assert d2 < d1


def test_setting_an_invalid_checksum(test_dataset):
    """Test whether setting an invalid checksum raises an error."""
    with pytest.raises(ValueError):
        test_dataset.md5sum = "invalid"


def test_checksum_verification(test_dataset):
    """Test whether setting an invalid checksum raises an error."""
    test_dataset.verify_checksum()
    test_dataset.md5sum = "0" * 32
    with pytest.raises(ValueError):
        test_dataset.verify_checksum()
